(* 
 * Copyright 2007 Robert Schneck <schneck@gmail.com>
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *)

(* The encoding triple *must* be three values which are *not* going to be
   garbage collected until after the custom model throws a signal
   (generally, row_deleted).

   Things may be more efficient if they are also "stable" (technically,
   not in the young generation for the garbage collector).  The C code
   will enforce this anyway by running a garbage collection whenever
   an unstable encoding is used.  The user may be able to ensure fewer 
   collections by replacing long-term row encodings with a "stable_copy" 
   before ever using it. *)
external stable_copy : 'a -> 'a = "ml_stable_copy_prime"

class virtual ['row,'a,'b,'c] custom_model_callback_object = object
  method virtual encode_iter : 'row -> 'a * 'b * 'c
  method virtual decode_iter : 'a -> 'b -> 'c -> 'row

  method custom_get_flags : GtkEnums.tree_model_flags list = []
  method virtual custom_get_n_columns : int
  method virtual custom_get_column_type : int -> Gobject.g_type
  method virtual custom_get_iter : Gtk.tree_path -> 'row option
  method virtual custom_get_path : 'row -> Gtk.tree_path
  method virtual custom_get_value : 'row -> int -> Gobject.g_value -> unit
  method virtual custom_iter_next : 'row -> 'row option
  method virtual custom_iter_children : 'row option -> 'row option
  method virtual custom_iter_has_child : 'row -> bool
  method virtual custom_iter_n_children : 'row option -> int
  method virtual custom_iter_nth_child : 'row option -> int -> 'row option
  method virtual custom_iter_parent : 'row -> 'row option
  method custom_ref_node (_:'row) : unit = ()
  method custom_unref_node (_:'row) : unit = ()
end

type _custom_model = [`custommodel|`treemodel] Gtk.obj

external create : unit -> _custom_model = "ml_custom_model_create"

external register_custom_model_callback_object : 
  _custom_model -> ('row,'a,'b,'c) #custom_model_callback_object -> unit = 
  "ml_register_custom_model_callback_object"
  
external custom_row_inserted : _custom_model -> Gtk.tree_path -> 'row -> unit =
    "ml_custom_model_row_inserted"
external custom_row_changed : _custom_model -> Gtk.tree_path -> 'row -> unit =
    "ml_custom_model_row_changed"
external custom_row_has_child_toggled : _custom_model -> Gtk.tree_path -> 'row -> unit =
    "ml_custom_model_row_has_child_toggled"
external custom_row_deleted : _custom_model -> 'row -> unit =
    "ml_custom_model_row_deleted"
external custom_rows_reordered : _custom_model -> Gtk.tree_path -> 'row option -> int array -> unit =
    "ml_custom_model_rows_reordered"

class virtual ['row,'a,'b,'c] custom_model obj = object (self)
  inherit GTree.model obj
  inherit ['row,'a,'b,'c] custom_model_callback_object
    
  initializer register_custom_model_callback_object obj self

  method connect = new GTree.model_signals obj
		     
  method custom_row_inserted (path:Gtk.tree_path) (iter:'row) : unit =
    custom_row_inserted obj path iter
  method custom_row_changed (path:Gtk.tree_path) (iter:'row) : unit =
    custom_row_changed obj path iter
  method custom_row_has_child_toggled (path:Gtk.tree_path) (iter:'row) : unit =
    custom_row_has_child_toggled obj path iter
  method custom_row_deleted (path:Gtk.tree_path) : unit =
    custom_row_deleted obj path
  method custom_rows_reordered (path:Gtk.tree_path) (iter_opt:'row option) (new_order:int array) : unit =
    custom_rows_reordered obj path iter_opt new_order
end

(**********************************************************************
 * Stuff which should be in LablGTK
 **********************************************************************)

external gtk_tree_view_set_enable_tree_lines : Gtk.widget Gtk.obj -> bool -> unit = "ml_gtk_tree_view_set_enable_tree_lines"
let tree_view_set_enable_tree_lines (view:GTree.view) enabled =
  let (a,b,_) = GMain.Main.version in
  if a >= 2 && b >= 10 then gtk_tree_view_set_enable_tree_lines view#as_widget enabled else ()

external gtk_tree_view_scroll_to_cell : Gtk.widget Gtk.obj -> Gtk.tree_path option -> Gtk.tree_view_column Gtk.obj option -> bool -> float -> float -> unit = "ml_bytecode_gtk_tree_view_scroll_to_cell" "ml_native_gtk_tree_view_scroll_to_cell"
let tree_view_scroll_to_cell (view:GTree.view) path_opt col_opt align_pair_opt =
  let col_opt = 
    begin match col_opt with
      | None -> None
      | Some c -> Some c#as_column
    end
  in
  begin match align_pair_opt with
    | None -> gtk_tree_view_scroll_to_cell view#as_widget path_opt col_opt false 0.0 0.0
    | Some (y,x) -> gtk_tree_view_scroll_to_cell view#as_widget path_opt col_opt true y x
  end

let tree_view_scroll_to_path view path =
  tree_view_scroll_to_cell view (Some path) None (Some (0.5,0.0))

let view_expand_to_path (view:GTree.view) (path:Gtk.tree_path) : unit =
  (* copied from gtk c source.  Luckily expand_row only signals if the row wasn't already expanded *)
  let depth = GTree.Path.get_depth path in
  let indices = GTree.Path.get_indices path in
  let path = GTree.Path.create [] in
  for i = 0 to (depth-2) do
    GTree.Path.append_index path indices.(i);
    view#expand_row path
  done

(**********************************************************************
 * LablGTK thread stuff---minor changes to what's already there
 **********************************************************************)

(* start Gtk; this is essentially gtkThInit.ml 
   = GtkThread.start () ... only we need to copy the gtk_thread_id_opt out for sync *)
let gtk_thread_id_opt : int option ref = ref None

let first_time = ref true

let start_gui() =
  begin match !first_time with
    | false -> ()
    | true ->
        ignore (GtkMain.Main.init ());
        first_time := false
  end;
  GtkThread.reset ();
  Thread.create begin fun () ->
    gtk_thread_id_opt := Some (Thread.id (Thread.self ()));
    GtkThread.main ();
    gtk_thread_id_opt := None
  end ()

(* Need to synchronize our messing with the model with the GTK thread.
   Oddly, GtkThread.sync doesn't actually seem to work. 
   It was fixed by version 2.6.0 (not yet in DarwinPorts?)
   This modified version does work. *)
let cannot_sync () =
  match !gtk_thread_id_opt with None -> true
  | Some id -> Thread.id (Thread.self ()) = id
let sync f =
  if cannot_sync () then begin f () end else begin
    let ready = ref false in
    let m = Mutex.create () in
    Mutex.lock m;
    let c = Condition.create () in
    GtkThread.async (fun () -> f (); ready := true; Condition.signal c) ();
    while not !ready do Condition.wait c m done
  end

(* change from LablGTK: if in GTK thread, don't bother with doing the
   thing asynchronously... just do it now.  This is because we only care
   about async to avoid thread bugs. *)
let async f = 
  if cannot_sync () then begin f () end else begin
    GtkThread.async f ()
  end

